#!/usr/bin/env python

"""
@copyright: Copyright (C) 2008, 2010 Oracle and/or its affiliates. All rights reserved.

@license:   See the file COPYING for redistribution information.

@summary:   Very simple interpreter used to test a Oracle VM Storage Connect
            plugins in isolation.
"""

import os
import sys

if "/opt/storage-connect" not in sys.path: sys.path.append("/opt/storage-connect")

# PyDev remote debugging setup
if os.environ.has_key('PYDEV_DEBUG') and os.path.exists('/opt/pydev/pysrc'):
    sys.path.append('/opt/pydev/pysrc')
    try:
        import pydevd #@UnresolvedImport
        if os.environ['PYDEV_DEBUG'].lower() in ['o', 'on', 'y',  'yes', 't', 'true', '1']:
            pydevd.settrace() 
    
        else:
            pydevd.settrace(os.environ['PYDEV_DEBUG'])

    except:
        pass

import cmd
import gettext
import optparse
import logging #@UnusedImport
import logging.handlers
import traceback

import OSCPlugin

from optparse import OptionParser
from PluginManager import PluginManager

_ = gettext.gettext

__revision__="15"

class PluginError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)

class PluginInstallError(PluginError):
    def __init__(self, msg):
        self.msg = msg

class PluginInvalidOption(PluginError):
    def __init__(self, msg):
        self.msg = msg

class ScriptError(Exception):
    def __init__(self, message):
        self.message = message
        self.args = message

    def __repr__(self):
        return self.message

    def __str__(self):
        return self.message

class TestFailed(Exception):
    def __init__(self, test, message):
        self.test = test
        self.message = message
        self.args = (test, message)

    def __repr__(self):
        return "%s: %s" % (self.test, self.message)

    def __str__(self):
        return "%s: %s" % (self.test, self.message)

class Interpreter(cmd.Cmd):
    def __init__(self, histfile):
        cmd.Cmd.__init__(self)
        self.prompt = "OSCPR>>> "
        self.intro = "Welcome to the Oracle Storage Connect Plugin Runner.\n\n" \
                      "Type help for more command information.\n\n"

        import readline
        import rlcompleter #@UnusedImport

        if histfile:
            try:
                readline.read_history_file(histfile)

            except IOError:
                pass

            import atexit
            atexit.register(readline.write_history_file, histfile)

        self._quit_ = False
        self._stdoutHandler_ = None
        self._indent_ = 0
        self._hist_ = []
        self._locals_ = {}
        self._globals_ = {
                          "getPlugin":      self.getPlugin,
                          "ScriptError":    ScriptError,
                          "TestFailed":     TestFailed,
                          }
        self._logger_ = None
        self._plugin_ = None
        self._plugin_mgr_ = PluginManager()

    def cmdloop(self, handle_exceptions = False, intro = None):
        self.preloop()
        if self.use_rawinput and self.completekey:
            try:
                import readline
                self.old_completer = readline.get_completer()
                readline.set_completer(self.complete)
                readline.parse_and_bind(self.completekey + ": complete")

            except ImportError:
                pass

        try:
            if intro is not None:
                self.intro = intro

            if self.intro:
                self.stdout.write(str(self.intro) + "\n")

            stop = None
            while stop is None:
                if self.cmdqueue:
                    line = self.cmdqueue.pop(0)

                else:
                    use_prompt = self.prompt
                    doexec = False
                    multiline = ''
                    if self.use_rawinput:
                        while not doexec:
                            try:
                                line = raw_input(use_prompt)
    
                            except EOFError:
                                line = 'EOF'

                            if not len(line) or line == 'EOF':
                                doexec = True

                            else:
                                if line[-1:] in ['\\', ':'] or len(multiline):
                                    if line[-1:] in ['\\', ':']:
                                        use_prompt = '.' * len(self.prompt)
                                        
                                    multiline += line + '\n'

                                else:
                                    doexec = True
                                    
                        if len(multiline):
                            line = multiline

                    else:
                        self.stdout.write(use_prompt)
                        self.stdout.flush()
                        while not doexec:
                            line = self.stdin.readline()
                            if not len(line):
                                doexec = True

                            else:
                                if line[-1:] in ['\\', ':'] or len(multiline):
                                    if line[-1:] in ['\\', ':']:
                                        use_prompt = '.' * len(self.prompt)
                                        
                                    multiline += line + '\n'

                                else:
                                    doexec = True
                                    
                        if len(multiline):
                            line = multiline

                line = self.precmd(line)
                try:
                    stop = self.onecmd(line)

                except Exception, e:
                    if handle_exceptions:
                        print "%s: %s" % (e.__class__.__name__, e)
                        continue

                    else:
                        raise

                stop = self.postcmd(stop, line)

            self.postloop()

        finally:
            if self.use_rawinput and self.completekey:
                try:
                    import readline
                    readline.set_completer(self.old_completer)

                except ImportError:
                    pass

    def get_names(self):
        # The cmd one is just plain broken!
        return dir(self)

    def completenames(self, text, line, begidx, endidx):
        if len(text) == 0:
            names = [a[3:] for a in dir(self)         if not a.startswith('_')] + \
                    [b     for b in self._globals_    if not b.startswith('_')] + \
                    [c     for c in dir(__builtins__) if not c.startswith('_')]

        elif text.rfind('.') > 0:
            dir_text = text[:text.rfind('.')]
            tab_text = text[text.rfind('.') + 1:]
            exec "__dirKW_list__ = dir(%s)" % dir_text in self._locals_, self._globals_
            if self._globals_.has_key("__dirKW_list__"):
                dir_list = self._globals_.pop("__dirKW_list__")
                if len(tab_text) == 0:
                    names = [dir_text + '.' + a for a in dir_list if not a.startswith('_')]

                else:
                    names = [dir_text + '.' + a for a in dir_list if a.startswith(tab_text)]

        else:
            names = [a[3:] for a in dir(self)         if a.startswith("do_" + text)] + \
                    [b     for b in self._globals_    if b.startswith(text)] + \
                    [c     for c in dir(__builtins__) if c.startswith(text)]

        return names

    def preloop(self):
        cmd.Cmd.preloop(self)
        if self._plugin_:
            self._globals_["plugin"] = self._plugin_

    def postloop(self):
        cmd.Cmd.postloop(self)
        if not self._quit_:
            print "\n\nExiting...\n"

    def precmd(self, line):
        self._hist_ += [line.strip()]
        return line

    def postcmd(self, stop, line):
        return stop

    def emptyline(self):
        pass

    def default(self, line):
        verbose = True
        try:
            if line.startswith('@'):
                if line.startswith('@@'):
                    verbose = False
                    if self._stdoutHandler_:
                        self._logger_.removeHandler(self._stdoutHandler_)

                    line = line[1:]

                else:
                    if not self._stdoutHandler_:
                        self._stdoutHandler_ = logging.StreamHandler()
                        self._stdoutHandler_.setLevel(logging.INFO)
                        self._logger_.addHandler(self._stdoutHandler_)

                self.do_script(line[1:])

                if self._stdoutHandler_:
                    self._logger_.removeHandler(self._stdoutHandler_)

            else:
                exec line in self._locals_, self._globals_

        except TestFailed, e:
            if self._stdoutHandler_:
                self._logger_.removeHandler(self._stdoutHandler_)

            self._logger_.fatal("%s" % e.__repr__())
            
            if not verbose:
                sys.stderr.write("\n\n")
                traceback.print_exception(sys.exc_type, sys.exc_value, sys.exc_traceback)
                
            else:
                raise
            
        except Exception, e:
            if self._stdoutHandler_:
                self._logger_.removeHandler(self._stdoutHandler_)

            self._logger_.fatal("%s: %s" % (e.__class__.__name__, e))
            
            if not verbose: 
                sys.stderr.write("\n\n")
                traceback.print_exception(sys.exc_type, sys.exc_value, sys.exc_traceback)

            else:
                raise

    def do_hist(self, args):
        """
        Print the list of commands that have been entered
        """
        print '\n'.join(self._hist_)

    def do_exit(self, args):
        """
        Exit from the interpreter
        """
        return True

    def do_quit(self, args):
        """
        Quit from the interpreter
        """
        self._quit_ = True
        return self._quit_

    def do_help(self, args):
        """
        Get help on commands
        'help' or '?' with no arguments prints a list of commands for which help
        is available
        'help <command>' or '? <command>' gives help on the specific <command>
        """
        if args == '@':
            args = "atsign"
        if args == "@@":
            args = "atatsign"
        cmd.Cmd.do_help(self, args)

    def do_EOF(self, args):
        """
        Exit from the interpreter
        """
        return True

    def do_script(self, line):
        """
        Execute a script (Same as @@)
        """
        line = line.strip()
        if len(line) > 0:
            if os.path.exists(line):
                file_path = os.path.abspath(line)
                file_name = os.path.basename(file_path)
                sys.modules["PluginRunner"] = sys.modules["__main__"]
                sys.modules["PluginRunner"].__name__ = "PluginRunner"

                from imp import new_module
                _mod = new_module("__main__")
                sys.modules["__main__"] = _mod
                _mod.__file__ = file_path
                globals = _mod.__dict__
                globals["SCRIPT_NAME"] = file_name
                if _mod.__file__.startswith(sys.path[0]):
                    del sys.path[0]

                sys.path.insert(0, os.path.split(file_path)[0])
                for gkey in self._globals_.keys():
                    globals[gkey] = self._globals_[gkey]

                __argv = sys.argv[:]
                sys.argv = [file_path]
                if self._globals_.get("argv", None) is not None:
                    sys.argv.append(self._globals_["argv"])

                execfile(file_path, globals)
                
                sys.argv = __argv
                
            else:
                print "Script file '%s' does not exist" % line

        else:
            print "*** No script name supplied"

    def help_atsign(self):
        """
        Execute script duplicating logs to terminal for monitoring
        """
        print "Execute script duplicating logs to terminal for monitoring"

    def help_atatsign(self):
        """
        Execute script without duplicating logging to terminal
        """
        print "Execute script without duplicating logging to terminal"

    def help_script(self):
        """
        Execute script without duplicating logging to terminal
        """
        print "Execute script without duplicating logging to terminal"

    def help_logger(self):
        """
        Log the text to the log file
        """
        exec "import logging\nhelp(logging)" in self._locals_, self._globals_

    def help_getPlugin(self):
        self.stdout.write("\n%s\n\n" %
                          "   Get the Plugin Manager class for the named plugin.\n\n"
                          "   Returns the Plugin Manager class for the named Storage Connect plugin.")

    def setLogger(self, logger):
        self._logger_ = logger
        self._globals_["logger"] = logger
        if self._plugin_mgr_:
            self._plugin_mgr_.setLogger(logger)

    def getPlugin(self, plugin_impl_name):
        if not self._logger_:
            raise PluginInstallError(_("No logger have been setup yet"))

        self._plugin_ = self._plugin_mgr_.getPlugin(plugin_impl_name)
        if not self._plugin_:
            raise PluginInstallError(_("Unable to locate plugin %s" % plugin_impl_name))

        return self._plugin_

    def setAdminHostName(self, admin_host):
        self._globals_["adminHostName"] = admin_host
        self._globals_["initialiseVariables"] = True

    def setAdminUserName(self, admin_user):
        self._globals_["adminUserName"] = admin_user
        self._globals_["initialiseVariables"] = True

    def setAdminPassword(self, admin_pass):
        self._globals_["adminPassword"] = admin_pass
        self._globals_["initialiseVariables"] = True

    def setExtraInfo(self, extra_info):
        self._globals_["extraInfo"] = extra_info
        self._globals_["initialiseVariables"] = True

    def setAccessHostName(self, access_host):
        self._globals_["accessHostName"] = access_host
        self._globals_["initialiseVariables"] = True

    def setUserName(self, user_name):
        self._globals_["userName"] = user_name
        self._globals_["initialiseVariables"] = True

    def setPassword(self, password):
        self._globals_["password"] = password
        self._globals_["initialiseVariables"] = True

    def setCHAPMode(self, chap_mode):
        self._globals_["CHAP"] = chap_mode

    def setStorageType(self, storage_type):
        self._globals_["storageType"] = storage_type
        self._globals_["initialiseVariables"] = True

    def setVolumeGroupName(self, vol_group_name):
        self._globals_["volGroupName"] = vol_group_name
        self._globals_["initialiseVariables"] = True

    def setPluginImplName(self, plugin_impl_name):
        self._globals_["pluginImplName"] = plugin_impl_name
        self._globals_["initialiseVariables"] = True

    def setScriptOptions(self, scriptOptions):
        self._globals_["argv"] = scriptOptions

    def setOKToRun(self, noQuestionsAsked):
        self._globals_["okToRun"] = noQuestionsAsked

class CmdLineOptions:
    def __init__(self, program_name, version):
        self.parser = OptionParser(usage = "%prog [options] ...",
                                   version = "Oracle VM Storage Connect "
                                             "Plugin Runner version %s" % version,
                                   prog = program_name,
                                   add_help_option = False)

        self.parser.disable_interspersed_args()

        self.parser.add_option("-h", "--help",
                               action = "callback",
                               callback = CmdLineOptions._help_opt_callback_,
                               callback_args = (self, "exit", 0))

        self.parser.add_option("--show-plugins",
                               dest = "show_plugins",
                               action = "store_true",
                               default = False,
                               help = "Display all the installed Storage Connect plugins on the"
                                      "system.")

        self.parser.add_option("--ignore-discover-errors",
                               dest = "ignore_discover_errors",
                               action = "store_true",
                               default = False,
                               help = optparse.SUPPRESS_HELP)

        self.parser.add_option("--plugin",
                               dest = "plugin_impl_name",
                               metavar = "IPlugin",
                               help = "Implementation name of the Storage Connect plugin to use")

        self.parser.add_option("--admin-host",
                               dest = "admin_host",
                               metavar = "adminHostName",
                               help = "Storage Server Administrator Hostname")

        self.parser.add_option("--admin-user",
                               dest = "admin_user",
                               metavar = "adminUserName",
                               help = "Storage Server Administration Username")

        self.parser.add_option("--admin-passwd",
                               dest = "admin_pass",
                               metavar = "adminPassword",
                               help = "Storage Server Administration Password")

        self.parser.add_option("--extra-info",
                               dest = "extra_info",
                               metavar = "extraInfo",
                               help = "Storage Server Extra Info field")

        self.parser.add_option("--access-host",
                               dest = "access_host",
                               metavar = "accessHostName",
                               help = "Storage Server iSCSI Access Hostname")

        self.parser.add_option("--username",
                               dest = "user_name",
                               metavar = "userName",
                               help = "Storage Server iSCSI Username")

        self.parser.add_option("--password",
                               dest = "password",
                               metavar = "password",
                               help = "Storage Server iSCSI Password")

        self.parser.add_option("--enable-chap",
                               dest = "chap",
                               metavar = "CHAP",
                               help = "Enable Storage Server iSCSI CHAP",
                               action = "store_true",
                               default = False)

        self.parser.add_option("--storage-type",
                               dest = "storage_type",
                               metavar = "storageType",
                               help = "Storage Server Storage Type (SAN/iSCSI/NFS/OCFS2)")

        self.parser.add_option("--vol-group-name",
                               dest = "vol_group_name",
                               metavar = "volumeGroupName",
                               help = "Volume group name to be used when creating LUNs")

        self.parser.add_option("--script",
                               dest = "script_name",
                               metavar = "scriptName",
                               help = "Non-interactive execution of script scriptName")

        self.parser.add_option("--script-options",
                               dest = "script_options",
                               metavar = "scriptOptions",
                               help = "Options to supply to the script (e.g. clean-first)")

        self.parser.add_option("--verbose",
                               dest = "verbose",
                               action = "store_true",
                               help = "In non-interactive script execution, the log output "
                                      "will be duplicated on the terminal.")

        self.parser.add_option("--log-dir",
                               dest = "log_dir",
                               default = ".",
                               metavar = "LogFileDir",
                               help = "Set the directory where the log file should be created"
                                      " (Defaults to .)")

        self.parser.add_option("--log-name",
                               dest = "log_file_name",
                               metavar = "LogFileName",
                               help = "Set the name of the log file "
                                      "(Defaults to pluginrunner_<implname>.log)")

        self.parser.add_option("--overwrite-log",
                               dest = "overwrite_log",
                               action = "store_true",
                               help = "Overwrite any old log file instead of"
                                      "appending to it.")

        self.parser.add_option("--yes-to-all",
                               dest = "yes_to_all",
                               action = "store_true",
                               help = "The tests are DESTRUCTIVE, this will allow tests to be run"
                                      "WITHOUT ASKING ANY confirmation.")

    @staticmethod
    def _help_opt_callback_(option, opt, value, parser, *xargs):
        xargs[0].printHelp()
        if xargs[1] == "exit":
            import sys #@Reimport
            sys.exit(xargs[2])

    def parse(self, args):
        (self.options, self.xtra_args) = self.parser.parse_args(args)
        return (self.options, self.xtra_args)

    def getParsedOptions(self):
        return (self.options, self.xtra_args)

    def printHelp(self):
        self.parser.print_help()
        print "\n"
        print "for example:\n"
        print "%s    \\" % self.parser.prog
        print "  --plugin=oracle.sample.SCSIPlugin.SamplePlugin \\"
        print "  --admin-host=myarray.example.com               \\"
        print "  --admin-user=admin                             \\"
        print "  --admin-passwd=password                        \\"
        print "  --storage-type=SAN                             \\"
        print "  --vol-group-name=MyVolGroup                    \\"
        print "  --script=MyTestScript.pscr                     \\"
        print "  --script-options=clean-first                   \\"
        print "  --log-dir=/tmp                                 \\"
        print "  --log-name=pluginrunner_sampleplugin.log       \\"
        print "  --overwrite-log\n"
        print "OR"
        print "%s --show-plugins" % self.parser.prog
        print "\n"

def main(args):
    program_name = os.path.basename(sys.argv[0])
    if program_name.rfind(".") > 0:
        program_name = program_name[:program_name.rfind(".")]

    version = ""
    for vers_sect in OSCPlugin.IPlugin.plugin_api_version:
        version = version + vers_sect + "."

    version = version[:-1] + __revision__
    histfile = os.path.join(os.environ["HOME"], ".%s_hist" % program_name.lower())

    try:
        cmd_line_opts = CmdLineOptions(program_name, version)
        options, xtra_args = cmd_line_opts.parse(args) #@UnusedVariable
        
        if options.show_plugins:
            pluginManager = PluginManager()
            print "\nInstalled Storage Connect plugins:\n"
            for pluginName in pluginManager.discoverPlugins(ignore_failed = options.ignore_discover_errors):
                print "    %s" % pluginName
            
            print "\n"
            sys.exit(0)

        if options.overwrite_log:
            mode = "w"

        else:
            mode = "a"

        if not options.log_file_name:
            if options.plugin_impl_name:
                name_pfxm, plugin_base_name = options.plugin_impl_name.rsplit('.', 1) #@UnusedVariable
                log_name = "%s/%s_%s.log" % (options.log_dir,
                                             program_name.lower(),
                                             plugin_base_name)

            else:
                log_name = "%s/%s.log" % (options.log_dir, program_name)

        else:
            log_name = "%s/%s" % (options.log_dir, options.log_file_name)

        log_format = logging.Formatter("<%(asctime)s %(levelname)s> %(message)s",
                                       "%y/%m/%d %H:%M:%S")

        logger = logging.getLogger(program_name)
        default_handler = logging.handlers.RotatingFileHandler(log_name,
                                                               mode = mode,
                                                               backupCount = 7)
        default_handler.setFormatter(log_format)
        logger.addHandler(default_handler)

        if options.verbose:
            logger.setLevel(logging.DEBUG)
            default_handler.setLevel(logging.DEBUG)
            
        else:
            logger.setLevel(logging.INFO)
            default_handler.setLevel(logging.INFO)

        interpreter = Interpreter(histfile)
        interpreter.setLogger(logger)
        logger.info("%s Start %s" % ('=' * 15, '=' * 15))
        if options.plugin_impl_name:
            if options.verbose:
                logger.info('pluginImplName = "%s"', options.plugin_impl_name)
            interpreter.setPluginImplName(options.plugin_impl_name)
            interpreter.getPlugin(options.plugin_impl_name)

        if options.admin_host:
            if options.verbose:
                logger.info('adminHostName  = "%s"', options.admin_host)
            interpreter.setAdminHostName(options.admin_host)

        if options.admin_user:
            if options.verbose:
                logger.info('adminUserName  = "%s"', options.admin_user)
            interpreter.setAdminUserName(options.admin_user)

        if options.admin_pass:
            if options.verbose:
                logger.info('adminPassword  = "********"')
            interpreter.setAdminPassword(options.admin_pass)

        if options.extra_info:
            if options.verbose:
                logger.info('extraInfo      = "%s"', options.extra_info)
            interpreter.setExtraInfo(options.extra_info)

        if options.access_host:
            if options.verbose:
                logger.info('accessHostName = "%s"', options.access_host)
            interpreter.setAccessHostName(options.access_host)

        if options.user_name:
            if options.verbose:
                logger.info('userName       = "%s"', options.user_name)
            interpreter.setUserName(options.user_name)

        if options.password:
            if options.verbose:
                logger.info('password       = "%s"', options.password)
            interpreter.setPassword(options.access_host)

        if options.chap is not None:
            if options.verbose:
                logger.info('CHAP           = "%s"', options.chap)
            interpreter.setCHAPMode(options.chap)

        if options.storage_type:
            if options.verbose:
                logger.info('storageType    = "%s"', options.storage_type)
            interpreter.setStorageType(options.storage_type)

        if options.vol_group_name:
            if options.verbose:
                logger.info('volGroupName   = "%s"', options.vol_group_name)
            interpreter.setVolumeGroupName(options.vol_group_name)

        if options.yes_to_all:
            if options.verbose:
                logger.info("okToRun        = True")
            interpreter.setOKToRun(options.yes_to_all)

        if options.script_options:
            if options.verbose:
                logger.info('scriptOptions  = "%s"', options.script_options)
            interpreter.setScriptOptions(options.script_options)

        if options.script_name:
            if options.verbose:
                logger.info('scriptName     = "%s"', options.script_name)
                interpreter.cmdqueue.append("@ %s" % options.script_name)

            else:
                interpreter.cmdqueue.append("@@ %s" % options.script_name)

            interpreter.cmdqueue.append("quit")

        logger.debug("Start interpreter...")
        if options.verbose:
            interpreter.cmdloop(False)
        
        else:
            interpreter.cmdloop(True)
            
        logger.info("%s End %s" % ('=' * 15, '=' * 15))

    except PluginError, e:
        sys.stderr.write("\nERROR: %s\n" % e.msg)
        sys.exit(1)

    except PluginInvalidOption, e:
        sys.stderr.write("\nERROR: %s\n" % e.msg)
        if cmd_line_opts:
            cmd_line_opts.printHelp()

        sys.exit(2)

#
# Entry point...
#
if __name__ == '__main__': main(sys.argv[1:])
